/*
			Copyright (c) 1995 by
			Advanced Visual Systems Inc.
			All Rights Reserved

	This software comprises unpublished confidential information of
	Advanced Visual Systems Inc. and may not be used, copied or made
	available to anyone, except in accordance with the license
	under which it is furnished.

	This file is under Perforce control
	$Id: //depot/express/fcs992/modules/rd_tbl.c#1 $
*/
#include <stdio.h>
#include <avs/f_utils.h>
#include <avs/util.h>
#include <avs/err.h>
#include <avs/om.h>
#include <avs/math.h>
#include <avs/dtype.h>
#include <avs/fld.h>

#define LINE_LEN 1024
#define MAX_COL 256
#define MAX_TYPES 6
#define ALLOC_BLOCK 1024

#define ERR_RETURN(A) ERRerror("DVread_table_read",1,ERR_ORIG, A); return(0);

static char *ui_types[MAX_TYPES] = {"byte","short","int","float","double","string"};
static char *formats[] = {"%c","%c","%h","%d","%f","%lf", " ", "%s"};

typedef struct _TBLdata {
	int use;
	int type;
	int size;
	int count;
	int null_flag;
	double null_value;
	char *values;
	char **strings;
	char *name;
} TBLdata;

static int get_type (char *token);
static int read_token (TBLdata *data, char *token);
static int free_data (TBLdata *data, int ndata);

int DVtable_read_init_update(OMobj_id elem_id, OMevent_mask event_mask, int seq_num)
{
	OMobj_id file, out, val_id; 
	char *filename, *delim, line[LINE_LEN], token[LINE_LEN], *names[MAX_COL], tmp[64], *ptr;
	int  i, j, count, ncol, ncol1, delim_len, line_len, found, type, types[MAX_COL];
	int  header, skip_lines, read_file;
	FILE     *pf;
	double null_value;

	if (OMget_name_int_val(elem_id, OMstr_to_name("init_file"), &read_file) != 1) 
		read_file = 0;
	
	if (read_file == 0) {
		out = OMfind_subobj(elem_id, OMstr_to_name("out"), OM_OBJ_RW);
		if (OMget_name_int_val(elem_id, OMstr_to_name("ncol"), &ncol) != 1)
			ncol = 0;

		for (i=0; i<ncol; i++) {
			if (OMget_array_val(out, i, &val_id, OM_OBJ_RW) != 1) {
				ERR_RETURN("cannot get out array member\n");
			}
			if (OMget_name_str_val(val_id, OMstr_to_name("name"), &ptr, 0) != 1) {
				sprintf(line, "column %d", i);
				if (OMset_name_str_val(val_id, OMstr_to_name("name"), line) != 1) {
					ERR_RETURN("cannot set column name\n");
				}
			}
			else free(ptr);
			if (OMget_name_int_val(val_id, OMstr_to_name("type"), &type) != 1) {
				if (OMset_name_int_val(val_id, OMstr_to_name("type"), 2) != 1) {
					ERR_RETURN("cannot set column type\n");
				}
			}
			if (OMget_name_int_val(val_id, OMstr_to_name("use"), &type) != 1) {
				if (OMset_name_int_val(val_id, OMstr_to_name("use"), 1) != 1) {
					ERR_RETURN("cannot set column type\n");
				}
			}
			if (OMget_name_real_val(val_id, OMstr_to_name("null_value"), &null_value) != 1) {
				if (OMset_name_real_val(val_id, OMstr_to_name("null_value"), 0.0) != 1) {
					ERR_RETURN("cannot set column type\n");
				}
			}
		}
		return(0);
	}

	filename = NULL;
	file = OMfind_subobj(elem_id, OMstr_to_name("filename"), OM_OBJ_RD);
	if (OMget_str_val(file, &filename, 0) != 1)
		return(0);
	if (filename[0] == '\0')
		return(1);
	if (OMget_name_str_val(elem_id, OMstr_to_name("delimiter"), &delim, 0) != 1)
		delim = strdup(" ");
	if (OMget_name_int_val(elem_id, OMstr_to_name("header"), &header) != 1) 
		header = 0;
	
	if (OMget_name_int_val(elem_id, OMstr_to_name("skip_lines"), &skip_lines) != 1) 
		skip_lines = 0;
	
	out = OMfind_subobj(elem_id, OMstr_to_name("out"), OM_OBJ_RW);
	if (OMset_name_int_val(elem_id, OMstr_to_name("ncol"), 0 ) != 1) {
		ERR_RETURN("cannot set ncol\n");
	}
	OMset_array_size(out, 0);

	if (!(pf = FILEfopen(filename, SIO_R_BIN))) {
		ERRerror("Read Table",1,ERR_ORIG, "Can't open data file: %s",
			 filename); 
		return(0);
	}
	for (i=0; i<skip_lines; i++) {
		if (fgets(line, LINE_LEN, pf) == NULL) {
			fclose(pf);
			ERR_RETURN("Error skipping lines");
		}
	}
	ncol = 0;
	delim_len = strlen(delim);

	/***  first line ***/
	/***  get names  ***/
	if (fgets(line, LINE_LEN, pf) == NULL) {
		fclose(pf);
		ERR_RETURN("Error reading header");
	}
	line_len = strlen(line);

	if (line[line_len-1] == '\r')
		line_len--;
	if (line[line_len-1] == '\n')
		line_len--;
	if (line[line_len-1] == '\r')
		line_len--;
		

	for (found=0,count = 0, i=0; i<line_len; i++) {
		if (line[i] == ' ' && !found)
			continue;   /* skip leading spaces */
		if (strncmp(line+i, delim, delim_len) == 0) {
			token[count++] = '\0';
			if (strlen(token) == 0)
				continue;

			if (header) {
				if (count>1)
					names[ncol++] = strdup(token);
				else {
					sprintf(tmp,"%d", ncol);
					strcpy(token, "column_");
					strcat(token, tmp);
					names[ncol++] = strdup(token);
				}
			}
			else {
				types[ncol] = get_type(token);
				sprintf(tmp,"%d", ncol);
				strcpy(token, "column_");
				strcat(token, tmp);
				names[ncol++] = strdup(token);
			}
			count = 0;
			i += delim_len-1;
			found = 0;
		}
		else {
			token[count++] = line[i];
			found = 1;
			if (i==line_len-1) {
				token[count++] = '\0';
				if (header) {
					names[ncol++] = strdup(token);
				}
				else {
					types[ncol] = get_type(token);
					sprintf(tmp,"%d", ncol);
					strcpy(token, "column_");
					strcat(token, tmp);
					names[ncol++] = strdup(token);
				}
			}
		}
	}
	ncol1 = ncol;
	if (header) {   /** get types from next line **/
		ncol = 0;
		if (fgets(line, LINE_LEN, pf) == NULL) {
			fclose(pf);
			ERR_RETURN("Error reading header");
		}
		line_len = strlen(line);

		if (line[line_len-1] == '\r')
			line_len--;
		if (line[line_len-1] == '\n')
			line_len--;
		if (line[line_len-1] == '\r')
			line_len--;
		
		for (found=0, count = 0, i=0; i<line_len; i++) {
			if (line[i] == ' ' && !found)
				continue;   /* skip leading spaces */
			if (strncmp(line+i, delim, delim_len) == 0) {
				token[count++] = '\0';
				count = 0;
				types[ncol++] = get_type(token);
				i += delim_len-1;
				found = 0;
			}
			else {
				token[count++] = line[i];
				found = 1;
				if (i==line_len-1) {
					token[count++] = '\0';
					types[ncol++] = get_type(token);
				}
			}
		}
		if (ncol != ncol1) {
			if (strcmp(delim, " ") == 0) {
				for (i=ncol; i<ncol1; i++) 
					types[ncol++] = get_type(" ");
			}
			else {
				fclose(pf);
				if (filename) 
					free(filename);
				if (delim)
					free(delim);
				ERR_RETURN("Number of columns found in first and second lines are different. Check column separator");
			}
		}
	}

	if (OMset_name_int_val(elem_id, OMstr_to_name("ncol"), ncol) != 1) {
		ERR_RETURN("cannot set ncol\n");
	}
	for (i=0; i<ncol; i++) {
		if (OMget_array_val(out, i, &val_id, OM_OBJ_RW) != 1) {
			ERR_RETURN("cannot get out array member\n");
		}
		if (OMset_name_str_val(val_id, OMstr_to_name("name"), names[i]) != 1) {
			ERR_RETURN("cannot set column name\n");
		}
		for (j=0; j<MAX_TYPES; j++) {
			if (DTYPEstr_to_type_id(ui_types[j]) == types[i]) {
				type = j;
				break;
			}
			if (j == MAX_TYPES-1) {
				type = j;
				ERRerror("DVread_tbl_header",1,ERR_ORIG, "unkown column type");
			}
		}
		if (OMset_name_int_val(val_id, OMstr_to_name("type"), type) != 1) {
			ERR_RETURN("cannot set column type\n");
		}
		if (OMset_name_int_val(val_id, OMstr_to_name("use"), 1) != 1) {
			ERR_RETURN("cannot set column set value\n");
		}
		if (OMset_name_real_val(val_id, OMstr_to_name("null_value"), 0.0) != 1) {
			ERR_RETURN("cannot set column type\n");
		}
	}
	
	
	fclose(pf);
	if (filename) 
		free(filename);
	if (delim)
		free(delim);
	for (i=0; i<ncol; i++)
		if (names[i])
			free(names[i]);
	return(1);
}

int DVtable_read_update(OMobj_id elem_id, OMevent_mask event_mask, int seq_num)
{
	OMobj_id file, in, out, val_id, e_id; 
	char *filename, *delim, line[LINE_LEN], token[LINE_LEN];
	int  i, j, stat, count, ncol, ndata, delim_len, line_len, line_count, col_count, found, type;
	int  header, skip_lines, read_file, in_ncol, nrow;
	TBLdata *data, *pdata;
	FILE     *pf;
	
	if (!(event_mask & OM_EVENT_INST)) {
		if (OMget_name_int_val(elem_id, OMstr_to_name("do"), &read_file) != 1) 
			read_file = 0;
		
		if (read_file == 0)
			return(0);
		else
			if (OMset_name_int_val(elem_id, OMstr_to_name("do"), 0) != 1) {
				ERR_RETURN("cannot set do\n");
			}
	}
	
	filename = NULL;
	file = OMfind_subobj(elem_id, OMstr_to_name("filename"), OM_OBJ_RD);
	if (OMget_str_val(file, &filename, 0) != 1)
		return(0);
	if (filename[0] == '\0')
		return(1);
	if (OMget_name_str_val(elem_id, OMstr_to_name("delimiter"), &delim, 0) != 1)
		delim = strdup(" ");
	if (OMget_name_int_val(elem_id, OMstr_to_name("header"), &header) != 1) 
		header = 0;
	
	if (OMget_name_int_val(elem_id, OMstr_to_name("skip_lines"), &skip_lines) != 1) 
		skip_lines = 0;
	
	in = OMfind_subobj(elem_id, OMstr_to_name("columns"), OM_OBJ_RD);
	OMget_array_size(in, &in_ncol);

	if (OMset_name_int_val(elem_id, OMstr_to_name("tbl_nrow"), 0) != 1) {
		ERR_RETURN("cannot set nrow\n");
	}

	if (OMset_name_int_val(elem_id, OMstr_to_name("tbl_ncol"), 0) != 1) {
		ERR_RETURN("cannot set ncol\n");
	}

	out = OMfind_subobj(elem_id, OMstr_to_name("out"), OM_OBJ_RW);
	OMset_array_size(out, 0);
	if (in_ncol == 0)
		return(0);

	data = (TBLdata *)malloc(in_ncol*sizeof(TBLdata));
				 
	for (ndata=0,i=0; i<in_ncol; i++) {
		pdata = data+i;
		if (OMget_array_val(in, i, &val_id, OM_OBJ_RD) != 1) {
			ERR_RETURN("cannot get in array member\n");
		}
		if (OMget_name_int_val(val_id, OMstr_to_name("use"), &j) != 1)
			j = 0;
		if (j == 0) {
			pdata->use = 0;
			continue;
		}
		pdata->use = 1;

		if (OMget_name_str_val(val_id, OMstr_to_name("name"), &(pdata->name), 0) != 1) {
			ERR_RETURN("cannot get column name\n");
		}
		if (OMget_name_int_val(val_id, OMstr_to_name("type"), &type) != 1) {
			ERR_RETURN("cannot get column type\n");
		}
		type = DTYPEstr_to_type_id(ui_types[type]);
		pdata->type = type;
		if (type != DTYPE_STRING) 
			if (OMget_name_real_val(val_id, OMstr_to_name("null_value"), &(pdata->null_value)) != 1) 
				pdata->null_value = 0;
		
		pdata->null_flag = 0;
		pdata->count = 0;
		pdata->size = ALLOC_BLOCK;
		if (type != DTYPE_STRING) {
			pdata->values = ARRalloc(NULL, type, pdata->size, NULL);
			if (pdata->values == NULL) {
				ERR_RETURN("cannot allocate data");
			}
		}
		else {
			pdata->strings = (char **)malloc(pdata->size*sizeof(char*));
			if (pdata->strings == NULL) {
				ERR_RETURN("cannot allocate data");
			}
		}
		ndata++;
	}

	if (!(pf = FILEfopen(filename, SIO_R_BIN))) {
		ERRerror("Read Table",1,ERR_ORIG, "Can't open data file: %s",
			 filename); 
		return(0);
	}
	for (i=0; i<skip_lines; i++) {
		if (fgets(line, LINE_LEN, pf) == NULL) {
			fclose(pf);
			ERR_RETURN("Error skipping lines");
		}
	}
	delim_len = strlen(delim);

	if (header) {
		if (fgets(line, LINE_LEN, pf) == NULL) {
			fclose(pf);
			ERR_RETURN("Error reading header");
		}
	}
	
	
	OMset_array_size(out, ndata);

	line_count = 0;
	while (fgets(line, LINE_LEN, pf) != NULL) {
		line_len = strlen(line);

		if (line[line_len-1] == '\r')
			line_len--;
		
		if (line[line_len-1] == '\n')
			line_len--;
		
		if (line[line_len-1] == '\r')
			line_len--;
		col_count = 0;
		ncol = 0;
		for (found=0,count = 0, i=0; i<line_len; i++) {
			if (line[i] == ' ' && !found)
				continue;   /* skip leading spaces */
			if (strncmp(line+i, delim, delim_len) == 0) {
				token[count++] = '\0';
				if ((data+ncol)->use) {
					if (ncol >= in_ncol) {
						fclose(pf);
						if (filename) 
							free(filename);
						if (delim)
							free(delim);
						free_data(data, in_ncol);
						if (data)
							free((char *)data);
						ERRerror("DVread_table_read",4,ERR_ORIG,
							 "bad number of columns in file %s on line %d: %d instead %d", 
							 filename, line_count+1, ncol+1, ndata);
						return(0);
					}
					read_token(data+ncol, token);
					col_count++;
				}
				ncol++;
				count = 0;
				i += delim_len-1;
				found = 0;
			}
			else {
				token[count++] = line[i];
				found = 1;
				if (i==line_len-1) {
					token[count++] = '\0';
					if ((data+ncol)->use) {
						if (ncol >= in_ncol) {
							fclose(pf);
							if (filename) 
								free(filename);
							if (delim)
								free(delim);
							free_data(data, in_ncol);
							if (data)
								free((char *)data);
							ERRerror("DVread_table_read",4,ERR_ORIG,
								 "bad number of columns in file %s on line %d: %d instead %d", 
								 filename, line_count+1, ncol+1, ndata);
							return(0);
						}
						read_token(data+ncol, token);
						col_count++;
					}
				}
			}
		}
		if (col_count != ndata) {
			/* add null data to the rest of the columns */
			for (i=col_count; i<ndata; i++)
				read_token(data+i, "");
		}
		line_count++;
	}
	fclose(pf);
	if (filename) 
		free(filename);

	for (nrow=0, count=0, i=0; i<in_ncol; i++) {
		if ((data+i)->use == 0)
			continue;
		
		pdata = data+i;

		if (count) {
			if (nrow != pdata->count) {
				fclose(pf);
				if (filename) 
					free(filename);
				if (delim)
					free(delim);
				free_data(data, in_ncol);
				if (data)
					free((char *)data);
				ERR_RETURN("different number of rows in columns\n");
			}
			nrow = pdata->count;
		}
		else 
			nrow = pdata->count;

		if (OMget_array_val(out, count, &val_id, OM_OBJ_RW) != 1) {
			ERR_RETURN("cannot get out array member\n");
		}
		if (OMset_name_int_val(val_id, OMstr_to_name("nvals"), pdata->count) != 1) {
			ERR_RETURN("cannot set out nvals\n");
		}
		if (OMset_name_str_val(val_id, OMstr_to_name("labels"), pdata->name) != 1) {
			ERR_RETURN("cannot set out labels\n");
		}

		e_id = OMfind_subobj(val_id, OMstr_to_name("values"), OM_OBJ_RW);
		
		stat = OMset_data_type(e_id, pdata->type);
		if (stat != 1) {
			ERR_RETURN("cannot set out values type\n");
		}
		if (pdata->type == DTYPE_STRING) {
			for (j=0; j<pdata->count; j++) 
				if (OMset_str_array_val(e_id, j, pdata->strings[j]) != 1) {
					ERR_RETURN("cannot set string data\n");
				}
		}
		else {
			stat = OMset_array(e_id, pdata->type, pdata->values, pdata->count, OM_SET_ARRAY_FREE);
			if (stat != 1) {
				ERR_RETURN("cannot set out data values\n");
			}
		}

		if (OMset_name_int_val(val_id, OMstr_to_name("null_flag"), pdata->null_flag) != 1) {
			ERR_RETURN("cannot set out null_flag\n");
		}
		if (pdata->null_flag)
			if (OMset_name_real_val(val_id, OMstr_to_name("null_value"), pdata->null_value) != 1) {
				ERR_RETURN("cannot set out null_value\n");
			}
		count++;
	}
	if (OMset_name_int_val(elem_id, OMstr_to_name("tbl_nrow"), nrow) != 1) {
		ERR_RETURN("cannot set nrow\n");
	}
	if (OMset_name_int_val(elem_id, OMstr_to_name("tbl_ncol"), ndata) != 1) {
		ERR_RETURN("cannot set nrow\n");
	}
	if (delim)
		free(delim);

	free_data(data, in_ncol);

	if (data)
		free((char *)data);
	return(1);
}

static int get_type(char *token)
{
	int l,i,type, real_type;
	
	real_type = 0;
	l = strlen(token);
	for (i=0; i<l; i++) {
		if (token[i] == '\n')
			break;
		if ((token[i]>='0' && token[i]<='9') || token[i] == '+' || token[i] == '-' || token[i] == ' ')
			continue;
		else if (token[i] == '.' || token[i] =='E' || token[i] =='e')
			real_type = 1;
		else {
			type = DTYPE_STRING;
			return(type);
		}
	}
	if (real_type)
		type = DTYPE_DOUBLE;
	else
		type = DTYPE_INT;

	return(type);
}

static int read_token (TBLdata *data, char *token)
{
	char str[LINE_LEN];
	char *format;

	format = formats[data->type];

	if ( data->type != DTYPE_STRING && data->count >= data->size) {
		data->size += ALLOC_BLOCK;
		data->values = (char *)ARRrealloc(data->values, data->type, data->size, NULL);
		if (data->values == NULL) {
			ERR_RETURN("cannot allocate data");
		}
	}
	if (strlen(token) == 0) {
		switch (data->type) {
		      case DTYPE_BYTE:
		      case DTYPE_CHAR:
		{
			char *dvalues;
			
			dvalues = (char *)data->values;
			data->null_flag = 1;
			dvalues[(data->count)] = data->null_value;
			(data->count)++;
			break;
		}
		      case DTYPE_SHORT:
		{
			short *dvalues;
			
			dvalues = (short *)data->values;
			data->null_flag = 1;
			dvalues[(data->count)] = data->null_value;
			(data->count)++;
			break;
		}
		      case DTYPE_INT:
		{
			int *dvalues;
			
			dvalues = (int *)data->values;
			data->null_flag = 1;
			dvalues[(data->count)] = data->null_value;
			(data->count)++;
			break;
		}
		      case DTYPE_FLOAT:
		{
			float *dvalues;
			
			dvalues = (float *)data->values;
			data->null_flag = 1;
			dvalues[(data->count)] = data->null_value;
			(data->count)++;
			break;
		}
		      case DTYPE_DOUBLE:
		{
			double *dvalues;
			
			dvalues = (double *)data->values;
			data->null_flag = 1;
			dvalues[(data->count)] = data->null_value;
			(data->count)++;
			break;
		}
		      case DTYPE_STRING:
			if (data->count >= data->size) {
				data->size += ALLOC_BLOCK;
				data->strings = (char **)realloc(data->strings, sizeof(char *)*data->size);
				if (data->strings == NULL) {
					ERR_RETURN("cannot allocate data");
				}
			}
			data->strings[(data->count)] = strdup("");
			data->null_flag = 1;
			(data->count)++;
			break;
			
		      default:
			return (0);
			break;
		}
		return(1);
	}

	switch (data->type) {
	      case DTYPE_BYTE:
	      case DTYPE_CHAR: 
	{
		char *dvalues;
		
		dvalues = (char *)data->values;
		if (sscanf(token, format, dvalues+(data->count)) != 1) {
			data->null_flag = 1;
			dvalues[(data->count)] = data->null_value;
		}
		(data->count)++;
		break;
	}
	      case DTYPE_SHORT:
	{
		short *dvalues;
			
		dvalues = (short *)data->values;
		if (sscanf(token, format, dvalues+(data->count)) != 1) {
			data->null_flag = 1;
			dvalues[(data->count)] = data->null_value;
		}
		(data->count)++;

		break;
	}
	      case DTYPE_INT:
	{
		int *dvalues;

		dvalues = (int *)data->values;
		if (sscanf(token, format, dvalues+(data->count)) != 1) {
			data->null_flag = 1;
			dvalues[(data->count)] = data->null_value;
		}
		(data->count)++;

		break;
	}
	      case DTYPE_FLOAT:
	{
		float *dvalues;

		dvalues = (float *)data->values;
		if (sscanf(token, format, dvalues+(data->count)) != 1) {
			data->null_flag = 1;
			dvalues[(data->count)] = data->null_value;
		}
		(data->count)++;

		break;
	}
	      case DTYPE_DOUBLE:
	{
		double *dvalues;

		dvalues = (double *)data->values;
		if (sscanf(token, format, dvalues+(data->count)) != 1) {
			data->null_flag = 1;
			dvalues[(data->count)] = data->null_value;
		}
		(data->count)++;

		break;
	}
	      case DTYPE_STRING:
		if (data->count >= data->size) {
			data->size += ALLOC_BLOCK;
			data->strings = (char **)realloc(data->strings, sizeof(char *)*data->size);
			if (data->strings == NULL) {
				ERR_RETURN("cannot allocate data");
			}
		}
/* CFS case 15853 sscanf will chop off strings with spaces
		if (sscanf(token, format, str) != 1) {
			data->strings[(data->count)] = strdup("");
			data->null_flag = 1;
		}
		else
*/
		data->strings[(data->count)] = strdup(token);
		(data->count)++;
		break;
		
	      default:
		return (0);
		break;
	}
	return(1);
}

static int free_data (TBLdata *data, int ndata)
{
	int i, j;
	TBLdata *pd;
	
	for (i=0; i<ndata; i++) {
		pd = data+i;
		if (pd->use == 0)
			continue;
		if (pd->type == DTYPE_STRING)  {
			for (j=0; j<pd->count; j++)
				free(pd->strings[j]);
			free(pd->strings);
		}
		free(pd->name);
	}
	return(1);
}

